#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Copyright (c) 2012, Nathan Sheldon. All rights reserved.
# http://www.nathansheldon.com/files/Dimmer-Extender.php
#
# Version 1.1.1
#
#		1.1.1 (13-Aug-2013)
#		* Fixed a bug that updated Indigo with the wrong actual brightness of
#		  a dimmer after Dimmer Extender sent a Set Brightness with Ramp Rate
#		  command to it.
#		--
#		1.1.0 (12-Aug-2013)
#		* Added Brighten by 1 Step and Dim by 1 Step actions which can be sent
#		  to any INSTEON device.
#		* Added Start Brightening, Start Dimming, and Stop Brightening/Dimming
#		  actions which can be sent to any INSTEON device.
#		--
#		1.0.2 (30-Jul-2013)
#		* Added the indigoPluginUpdateChecker module (code by Travis CooK)
#		  to facilitate automatic plugin version checking.
#		--
#		1.0.1 (14-Nov-2012)
#		* Fixed a bug that would cause the Indigo 6 beta client to crash
#		  during action validation.
#		--
#		1.0 (08-Nov-2012)
#		* Updated actions and dialogs for Indigo 6 UI.
#		* Added triggerable events. Dimmer Extender now generates
#		  events for INSTEON Start Brightening, INSTEON Stop Brightening,
#		  INSTEON Start Dimming, and INSTEON Stop Dimming. These events
#		  can be selected in any Indigo Trigger configuration window.
#		--
#		0.9.2 (04-Oct-2012):
#		* Fixed bug that would incorrectly report an error if a ramp rate
#		  variable was specified in the UI configuration.
#		* Fixed some error message typos.
#		--
#		0.9.1 (01-Oct-2012):
#		* Moved the command sending delay code into the runConcurrentThread
#		  method.  This will cause messages marked for delay to delay all
#		  subsequently added commands to be delayed, but will reduce signal
#		  collision issues during high INSTEON traffic conditions.
#		* Changed logging behavior so that the "Sent INSTEON" log entry
#		  is not reported until confirmation of the command's success has
#		  been received.  If a device was set to 0 percent brightness, the
#		  log now shows "...off at ramp rate..." rather than "...on to
#		  <brightness level> at ramp rate...".  The log error type, should
#		  the command fail, has been changed from "INSTEON Send Error" to
#		  "INSTEON Error" for slightly better readability.
#		* Added a few more lines of debug logging code.
#		* Switched to using the "time" module's version of the "sleep"
#		  method to avoid hangs when the "sleep" method is called while
#		  the runConcurrentThread is waiting for an INSTEON acknowledgement.
#		--
#		0.9 (03-Sep-2012):
#		* Initial beta release.


################################################################################
# Imports
from math import ceil
import indigoPluginUpdateChecker

################################################################################
# Globals
################################################################################
presetRates = [480.0, 360.0, 270.0, 210.0, 150.0, 90.0, 47.0, 38.5, 32.0, 28.0, 23.5, 19.0, 6.5, 2.0, 0.3, 0.1]
presetBrightnesses = [100, 94, 87, 81, 75, 69, 62, 56, 50, 44, 37, 31, 25, 18, 12, 0]

# Since Python 2.5 doesn't include the "bin()" method to convert
#   integers to binary strings, the below method duplicates the
#   functionality found in bin(). This code was taken from
#   http://code.activestate.com/recipes/219300-format-integer-as-binary-string/
try:
    if bin(0): pass
except NameError, ne:
	def bin(x):
		"""
		bin(number) -> string

		Stringifies an int or long in base 2.
		"""
		if x < 0: return '-' + bin(-x)
		out = []
		if x == 0: out.append('0')
		while x > 0:
			out.append('01'[x & 1])
			x >>= 1
			pass
		try: return '0b' + ''.join(reversed(out))
		except NameError, ne2: out.reverse()
		return '0b' + ''.join(out)

# Trigger Event Constants.
kTriggerType_StartBrightening = "startBrightening"
kTriggerType_StopBrightening = "stopBrightening"
kTriggerType_StartDimming = "startDimming"
kTriggerType_StopDimming = "stopDimming"

################################################################################
class Plugin(indigo.PluginBase):
	########################################
	# Class properties
	########################################
	commandQueue = []
	
	########################################
	# Called by Indigo when the plugin is loaded.
	def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs): 
		indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
		self.debug = pluginPrefs.get('showDebugInfo', False)
		self.debugLog("Initializing Plugin.")
		self.events = dict()
		self.events[kTriggerType_StartBrightening] = dict()
		self.events[kTriggerType_StopBrightening] = dict()
		self.events[kTriggerType_StartDimming] = dict()
		self.events[kTriggerType_StopDimming] = dict()
		# Include update checker module.
		self.updater = indigoPluginUpdateChecker.updateChecker(self, 'http://www.nathansheldon.com/files/PluginVersions/Dimmer-Extender.html')
		# Subscribe to incoming INSTEON messages.
		indigo.insteon.subscribeToIncoming()
	
	########################################
	# Called by Indigo when the plugin is unloaded.
	def __del__(self):
		indigo.PluginBase.__del__(self)
	
	########################################
	# Device Start (needed for update check).
	def startup(self):
		self.debugLog(u"startup called.")
		self.updater.checkVersionPoll()
		
	########################################
	# Called by Indigo when a trigger based on a Dimmer Extender event is created.
	def triggerStartProcessing(self, trigger):
		self.debugLog("Start processing trigger " + str(trigger.id))
		triggerType = trigger.pluginTypeId
		if (triggerType == kTriggerType_StartBrightening) or (triggerType == kTriggerType_StopBrightening) or (triggerType == kTriggerType_StartDimming) or (triggerType == kTriggerType_StopDimming):
			self.events[triggerType][trigger.id] = trigger
		self.debugLog("Start trigger processing list: " + str(self.events))
	
	########################################
	# Called by Indigo when a trigger based on a Dimmer Extender event is deleted.
	def triggerStopProcessing(self, trigger):
		self.debugLog("Stop processing trigger " + str(trigger.id))
		triggerType = trigger.pluginTypeId
		if triggerType in self.events:
			if (triggerType == kTriggerType_StartBrightening) or (triggerType == kTriggerType_StopBrightening) or (triggerType == kTriggerType_StartDimming) or (triggerType == kTriggerType_StopDimming):
				if trigger.id in self.events[triggerType]:
					del self.events[triggerType][trigger.id]
		self.debugLog("Stop trigger processing list: " + str(self.events))
	
	#########################################
	# Called by Indigo when an INSTEON command is received if we've subscribed
	#   to incoming commands (which we did in the __init__ method).
	def insteonCommandReceived(self, command):
		self.debugLog(u"INSTEON Command Received: " + str(command))
		address = str(command.address)
		address = address.strip().upper()
		function = command.cmdFunc
		scene = command.cmdScene
		# If this was a start or stop of brightening or dimming, see what triggers we have.
		#   Start Brightening
		if function == "start brighten":
			for trigger in self.events[kTriggerType_StartBrightening].values():
				deviceId = int(trigger.pluginProps.get('deviceId', 0))
				triggerScene = int(trigger.pluginProps.get('buttonGroup', 1))
				deviceAddress = indigo.devices[deviceId].address
				if (deviceAddress == address) and (triggerScene == scene):
					indigo.trigger.execute(trigger)
		#   Stop Brightening
		if function == "stop brighten":
			for trigger in self.events[kTriggerType_StopBrightening].values():
				deviceId = int(trigger.pluginProps.get('deviceId', 0))
				triggerScene = int(trigger.pluginProps.get('buttonGroup', 1))
				deviceAddress = indigo.devices[deviceId].address
				if (deviceAddress == address) and (triggerScene == scene):
					indigo.trigger.execute(trigger)
		#   Start Dimming
		if function == "start dim":
			for trigger in self.events[kTriggerType_StartDimming].values():
				deviceId = int(trigger.pluginProps.get('deviceId', 0))
				triggerScene = int(trigger.pluginProps.get('buttonGroup', 1))
				deviceAddress = indigo.devices[deviceId].address
				if (deviceAddress == address) and (triggerScene == scene):
					indigo.trigger.execute(trigger)
		#   Stop Dimming
		if function == "stop dim":
			for trigger in self.events[kTriggerType_StopDimming].values():
				deviceId = int(trigger.pluginProps.get('deviceId', 0))
				triggerScene = int(trigger.pluginProps.get('buttonGroup', 1))
				deviceAddress = indigo.devices[deviceId].address
				if (deviceAddress == address) and (triggerScene == scene):
					indigo.trigger.execute(trigger)
	
	########################################
	# Called by Indigo when the plugin configuration dialog is closed.
	def closedPrefsConfigUi(self, valuesDict, userCancelled):
		if not userCancelled:
			self.debug = valuesDict.get("showDebugInfo", False)
			if self.debug:
				indigo.server.log("Debug logging enabled")
			else:
				indigo.server.log("Debug logging disabled")
	
	########################################
	# Called by Indigo when the Save button in an event config dialog is clicked.
	def validateEventConfigUi(self, valuesDict, typeId, triggerId):
		trigger = None
		triggerName = u"(new trigger)"
		deviceId = valuesDict.get('deviceId', "")
		if triggerId != 0:	# 0 = new trigger
			trigger = indigo.triggers[triggerId]
			triggerName = trigger.name
		self.debugLog(u"Validating trigger event config for type: " + typeId + u", trigger name: " + triggerName)
		self.debugLog(u"Trigger data:\n" + str(trigger))
		errorsDict = indigo.Dict()
		descString = u""
		if deviceId != "":
			device = indigo.devices[int(deviceId)]
			# Update the event description information.
			if typeId == "starBrightening":
				descString = u"received INSTEON start brightening from " + device.name
			elif typeId == "stopBrightening":
				descString = u"receivded INSTEON stop brightening from " + device.name
			elif typeId == "startDimming":
				descString = u"received INSTEON start dimming from " + device.name
			elif typeId == "stopDimming":
				descString = u"received INSTEON stop dimming from " + device.name
		else:
			errorsDict['deviceId'] = u"Please select a device."
			errorsDict['showAlertText'] = errorsDict['deviceId'] + "\n\n"
		
		valuesDict['description'] = descString
		if len(errorsDict) > 0:
			errorsDict['showAlertText'] = errorsDict['showAlertText'].strip()
			return (False, valuesDict, errorsDict)
		
		return (True, valuesDict)
	
	########################################
	# Called by Indigo when the Save button in an Action config dialog is clicked.
	def validateActionConfigUi(self, valuesDict, typeId, devId):
		device = indigo.devices[devId]
		self.debugLog(u"Validating action config for type: " + typeId + u", device: " + device.name)
		isError = False
		errorsDict = indigo.Dict()
		errorsDict['showAlertText'] = ""
		descString = u""
		if typeId == "setBrightness":
			brightnessSource = valuesDict.get('brightnessSource', False)
			brightness = valuesDict.get('brightness', False)
			brightnessVarId = valuesDict.get('brightnessVariable', False)
			brightnessDevId = valuesDict.get('brightnessDevice', False)
			delay = valuesDict.get('delay', False)
			useRateVariable = valuesDict.get('useRateVariable', False)
			rate = valuesDict.get('rate', False)
			rateVarId = valuesDict.get('rateVariable', False)
			retries = valuesDict.get('retries', False)
			
			if not brightnessSource:
				isError = True
				errorsDict['brightnessSource'] = u"Please specify a Brightness Source."
				errorsDict['showAlertText'] += errorsDict['brightnessSource'] + "\n\n"
			elif brightnessSource == "custom":
				if not brightness:
					isError = True
					errorsDict['brightness'] = u"Please specify a brightness level."
					errorsDict['showAlertText'] += errorsDict['brightness'] + "\n\n"
				else:
					try:
						brightness = int(brightness)
						if brightness < 0 or brightness > 100:
							isError = True
							errorsDict['brightness'] = u"Brightness levels must be a number between 0 and 100."
							errorsDict['showAlertText'] += errorsDict['brightness'] + "\n\n"
					except ValueError:
						isError = True
						errorsDict['brightness'] = u"Brightness levels must be a number between 0 and 100."
						errorsDict['showAlertText'] += errorsDict['brightness'] + "\n\n"
				descString += u"set brightness of \"" + device.name + "\" to " + str(brightness) + "%"
			elif brightnessSource == "variable":
				if not brightnessVarId:
					isError = True
					errorsDict['brightnessVariable'] = u"Please specify a variable to use for brightness level."
					errorsDict['showAlertText'] += errorsDict['brightnessVariable'] + "\n\n"
				else:
					try:
						brightnessVar = indigo.variables[int(brightnessVarId)]
						descString += u"set brightness of \"" + device.name + "\" to value in variable \"" + brightnessVar.name + "\""
					except IndexError:
						isError = True
						errorsDict['brightnessVariable'] = u"The specified variable does not exist in the Indigo database. Please choose a different variable."
						errorsDict['showAlertText'] += errorsDict['brightnessVariable'] + "\n\n"
			elif brightnessSource == "dimmer":
				if not brightnessDevId:
					isError = True
					errorsDict['brightnessDevice'] = u"Please specify a dimmer device to use as the source for brightness level."
					errorsDict['showAlertText'] += errorsDict['brightnessDevice'] + "\n\n"
				else:
					try:
						brightnessDev = indigo.devices[int(brightnessDevId)]
						descString += u"set brightness of \"" + device.name + "\" to current brightness of \"" + brightnessDev.name + "\""
					except IndexError:
						isError = True
						errorsDict['brightnessDevice'] = u"The specified device does not exist in the Indigo database. Please choose a different device."
						errorsDict['showAlertText'] += errorsDict['brightnessDevice'] + "\n\n"
					if brightnessDev.id == device.id:
						isError = True
						errorsDict['brightnessDevice'] = u"You cannot select the same dimmer as the one for which you're setting the brightness."
						errorsDict['showAlertText'] += errorsDict['brightnessDevice'] + "\n\n"
			
			if not useRateVariable:
				if not rate:
					isError = True
					errorsDict['rate'] = u"Please select a Ramp Rate."
					errorsDict['showAlertText'] += errorsDict['rate'] + "\n\n"
				else:
					try:
						rate = float(rate)
						if rate < 0.1 or rate > 480:
							isError = True
							errorsDict['rate'] = u"Ramp Rate times must be between 0.1 and 480 seconds."
							errorsDict['showAlertText'] += errorsDict['rate'] + "\n\n"
					except ValueError:
						isError = True
						errorsDict['rate'] = u"Ramp Rates must be between 0.1 and 480 seconds and cannot contain non-numeric characters."
						errorsDict['showAlertText'] += errorsDict['rate'] + "\n\n"
				descString += u" using ramp rate " + str(rate) + " sec"
			else:
				if not rateVarId:
					isError = True
					errorsDict['rateVariable'] = u"Please select a variable to use for the ramp rate."
					errorsDict['showAlertText'] += errorsDict['rateVariable'] + "\n\n"
				else:
					try:
						rateVar = indigo.variables[int(rateVarId)]
						descString += u" using ramp rate in variable \"" + rateVar.name + "\""
					except IndexError:
						isError = True
						errorsDict['rateVariable'] = u"The specified variable does not exist in the Indigo database. Please choose a different variable."
						errorsDict['showAlertText'] += errorsDict['rateVariable'] + "\n\n"
			
			if not delay:
				delay = 0
			else:
				try:
					delay = float(delay)
					if delay < 0 or delay > 1:
						isError = True
						errorsDict['delay'] = u"Delay values must be between 0 and 1 seconds."
						errorsDict['showAlertText'] += errorsDict['delay'] + "\n\n"
					else:
						descString += u" with a " + str(delay) + " sec delay"
				except ValueError:
					isError = True
					errorsDict['delay'] = u"Delay values must be between 0 and 1 seconds and cannot contain non-numeric characters."
					errorsDict['showAlertText'] += errorsDict['delay'] + "\n\n"

			if not retries:
				retries = 0
			else:
				try:
					retries = int(retries)
					if retries < 0 or retries > 10:
						isError = True
						errorsDict['retries'] = u"Retries value must be between 0 and 10."
						errorsDict['showAlertText'] += errorsDict['retries'] + "\n\n"
					else:
						descString += u", retrying " + str(retries)
						if retries == 1:
							descString += u" time"
						else:
							descString += u" times"
						descString += u" on error"
				except ValueError:
					isError = True
					errorsDict['retries'] = u"Retries value must be between 0 and 10 and cannot contain non-numeric characters."
					errorsDict['showAlertText'] += errorsDict['retries'] + "\n\n"
		
		valuesDict['description'] = descString
		if isError:
			errorsDict['showAlertText'] = errorsDict['showAlertText'].strip()
			return (False, valuesDict, errorsDict)
		
		return (True, valuesDict)
	
	########################################
	# Called by Indigo in a separate thread shortly after plugin startup is complete.
	def runConcurrentThread(self):
		self.debugLog("runConcurrentThread called.")
		loopCount = 0	# Used to avoid calling update check too often.
		#
		# Continuously check the command queue and send
		#   them as they come in.
		#
		try:
			while True:
				# Check for newer plugin versions.
				if loopCount % 100 == 0:
					# Only check every 10 seconds.
					self.updater.checkVersionPoll()
					# Reset the loop count.
					loopCount = 0
				# Increment the loop count.
				loopCount += 1
				
				self.sleep(0.1)	# Sleep just a bit.
				# If there's items in the queue, let's get to work.
				if len(self.commandQueue) > 0:
					self.debugLog("runConcurrentThread: Detected " + str(len(self.commandQueue)) + " INSTEON commands waiting to be sent.")
					# Start from the beginning of the queue.
					commandItems = self.commandQueue[0]
					# Extract the data from the item in the queue.
					id = commandItems[0]
					device = indigo.devices[id]
					command1 = commandItems[1]
					command2 = commandItems[2]
					brightness = commandItems[3]
					rate = commandItems[4]
					delay = commandItems[5]
					retries = commandItems[6]
					message = [command1, command2]
					# Remove the item from the queue.
					self.debugLog("runConcurrentThread: Processing command for \"" + device.name + "\". Removing command from queue.")
					self.commandQueue.remove(commandItems)
					
					# Sleep for the specified delay time (using the "time" module version of sleep).
					if delay > 0:
						self.debugLog("runConcurrentThread: Sleeping for delay time " + str(delay) + " seconds.")
						self.sleep(delay)
					
					# Try to send the command. Wait for a reply.
					while retries >= 0:
						try:
							self.debugLog("runConcurrentThread: Sending command to \"" + device.name + "\". Waiting for acknowledgement.")
							reply = indigo.insteon.sendRaw(device.address, message, waitUntilAck=True, suppressLogging=True)
							self.debugLog("    reply: " + str(reply))
							# Log the changes.
							if reply.cmdSuccess == False:
								if brightness > 0:
									errorMessage = u"\"" + device.name + u"\" on to " + str(brightness) + u" at ramp rate " + str(rate) + u" sec; send failed (no acknowledgment)."
								else:
									errorMessage = u"\"" + device.name + u"\" off at ramp rate " + str(rate) + u" sec; send failed (no acknowledgment)."
								if retries > 0:
									errorMessage += u" Retries remaining: " + str(retries)
								indigo.server.log(errorMessage, "INSTEON", isError=True)
							else:
								if brightness > 0:
									indigo.server.log(u"\"" + device.name + u"\" on to " + str(brightness) + u" at ramp rate " + str(rate) + u" sec.", u"Sent INSTEON")
								else:
									indigo.server.log(u"\"" + device.name + u"\" off at ramp rate " + str(rate) + u" sec.", u"Sent INSTEON")
								# Update the device state on the server.
								device.updateStateOnServer('brightnessLevel', brightness)
								# Send successful. No more retries needed.
								retries = -1
						except Exception, e:
							if brightness > 0:
								errorMessage = u"\"" + device.name + u"\" on to " + str(brightness) + " at ramp rate " + str(rate) + u" sec; send failed: " + str(e) + "."
							else:
								errorMessage = u"\"" + device.name + u"\" off at ramp rate " + str(rate) + u" sec; send failed: " + str(e) + "."
							if retries > 0:
								errorMessage += u" Retries remaining: " + str(retries)
							indigo.server.log(errorMessage, isError=True)
							pass
						# Decrement the number of retries.
						retries -= 1
		
		except self.StopThread:
			self.debugLog("runConcurrentThread stopped.")
			pass
		
		self.debugLog("runConcurrentThread exiting.")
	
	########################################
	# ACTION METHODS
	########################################
	
	# Set Brightness
	########################################
	def setBrightness(self, action):
		self.debugLog("setBrightness called.")
		
		command1 = "0x2e"	# Turn On with brightness and rate.
							# (0x2F is Turn Off with rate, brightness ignored).
		command2 = "0xff"	# Defaults to full brightness at 0.1 sec. rate.
		
		device = indigo.devices[action.deviceId]
		self.debugLog("Device: " + device.name + " (ID " + str(device.id) + ")")
		isDeviceI2CS = False	# i2CS devices may need to use extended commands.
		brightnessSource = action.props.get('brightnessSource', False)
		brightness = action.props.get('brightness', False)
		brightnessVarId = action.props.get('brightnessVariable', False)
		brightnessDevId = action.props.get('brightnessDevice', False)
		useRateVariable = action.props.get('useRateVariable', False)
		rate = action.props.get('rate', False)
		rateVarId = action.props.get('rateVariable', False)
		delay = action.props.get('delay', False)
		retries = action.props.get('retries', False)
		
		# Validate the action properties.
		if not brightnessSource:
			# The dimmer level source wasn't specified. Try to figure out
			#   the intended source based on passed data in the action call.
			if brightness.__class__ != bool:
				brightnessSource = "custom"
			elif brightnessVarId:
				brightnessSource = "variable"
			elif brightnessDevId:
				brightnessSource = "dimmer"
			else:
				self.errorLog(u"No brightness source information was provided.")
				return None
		
		if brightnessSource == "custom":
			if brightness == False and brightness.__class__ != int:
				self.errorLog(u"No brightness level was specified.")
				return None
			else:
				try:
					brightness = int(brightness)
					if brightness < 0 or brightness > 100:
						self.errorLog(u"Brightness level " + str(brightness) + " is outside the acceptable range of 0 to 100.")
						return None
				except ValueError:
					self.errorLog(u"Brightness level \"" + str(brightness) + "\" is invalid. Brightness values can only contain numbers.")
					return None
			self.debugLog("Brightness (source: custom): " + str(brightness) + ", class: " + str(brightness.__class__))
		
		elif brightnessSource == "variable":
			if not brightnessVarId:
				self.errorLog(u"No variable containing the brightness level was specified.")
				return None
			else:
				try:
					brightnessVar = indigo.variables[int(brightnessVarId)]
					# Embedding float method inside int method allows for fractional
					#   data but just drops everything after the decimal.
					brightness = int(float(brightnessVar.value))
					if brightness < 0 or brightness > 100:
						self.errorLog(u"Brightness level " + str(brightness) + " found in variable \"" + brightnessVar.name + "\" is outside the acceptable range of 0 to 100.")
						return None
				except ValueError:
					self.errorLog(u"Brightness level \"" + str(brightnessVar.value) + "\" found in variable \"" + brightnessVar.name + "\" is invalid. Brightness values can only contain numbers.")
					return None
				except IndexError:
					self.errorLog(u"The specified variable (ID " + str(brightnessVarId) + ") does not exist in the Indigo database.")
					return None
			self.debugLog("Brightness (source: variable): " + str(brightness) + ", class: " + str(brightness.__class__))
		
		elif brightnessSource == "dimmer":
			if not brightnessDevId:
				self.errorLog(u"No dimmer was specified as the brightness level source.")
				return None
			else:
				# See if the submitted value is a device ID or a device name.
				try:
					brightnessDevId = int(brightnessDevId)
					# Value is a device ID number.
				except ValueError:
					try:
						brightnessDevId = indigo.devices[brightnessDevId].name
						# Value is a device name.
					except KeyError:
						self.errorLog(u"No device with the name \"" + str(brightnessDevId) + "\" could be found in the Indigo database.")
						return None
				try:
					brightnessDev = indigo.devices[brightnessDevId]
					brightness = int(brightnessDev.states.get('brightnessLevel', False))
					if brightness < 0 or brightness > 100:
						self.errorLog(u"Brightness level " + str(brightness) + " of device \"" + brightnessDev.name + "\" is outside the acceptable range of 0 to 100.")
						return None
				except ValueError:
					self.errorLog(u"The device \"" + brightnessDev.name + "\" does not have a brightness level. Please ensure that the device is a dimmer.")
					return None
				except KeyError:
					self.errorLog(u"The specified device (ID " + str(brightnessDevId) + ") does not exist in the Indigo database.")
					return None
			self.debugLog("Brightness (source: other dimmer): " + str(brightness) + ", class: " + str(brightness.__class__))
		
		else:
			self.errorLog(u"Unrecognized brightness source \"" + str(brightnessSource) + u"\". Valid brightness sources are \"custom\", \"variable\", and \"dimmer\".")
			return None
		
		if not useRateVariable:
			if not rate and rate.__class__ == bool:
				self.errorLog(u"No ramp rate was specified.")
				return None
			else:
				try:
					rate = float(rate)
					if rate < 0 or rate > 480:
						self.errorLog(u"Ramp rate value " + str(rate) + " is outside the acceptible range of 0 to 480.")
						return None
				except ValueError:
					self.errorLog(u"Ramp rate value \"" + str(rate) + " is an invalid value. Ramp rate values can only contain numbers.")
					return None
			self.debugLog("Rate: " + str(rate))
		
		else:
			if not rateVarId:
				self.errorLog(u"No variable containing the ramp rate time was specified.")
				return None
			else:
				try:
					rateVar = indigo.variables[int(rateVarId)]
					rate = float(rateVar.value)
					if rate < 0 or rate > 480:
						self.errorLog(u"Ramp rate value \"" + str(rate) + "\" found in variable \"" + rateVar.name + "\" is outside the acceptible range of 0 to 480.")
						return None
				except ValueError:
					self.errorLog(u"Ramp rate value \"" + str(rate) + "\" found in variable \"" + rateVar.name + "\" is an invalid value. Ramp rate values can only contain numbers.")
					return None
				except IndexError:
					self.errorLog(u"The specified variable (ID " + str(brightnessVarId) + ") does not exist in the Indigo database.")
			self.debugLog("Rate: " + str(rate))
		
		if not delay:
			delay = 0
		else:
			try:
				delay = float(delay)
				if delay < 0 or delay > 1:
					self.errorLog(u"Delay value \"" + str(delay) + "\" is outside the range of 0 to 1 second. Using default delay of 0 seconds.")
					delay = 0
			except ValueError:
				self.errorLog(u"Delay value \"" + str(delay) + "\" is not a valid number between 0 and 1 second. Using default delay of 0 seconds.")
				delay = 0
		self.debugLog("Delay: " + str(delay))
		
		if not retries:
			retries = 0
		else:
			try:
				retries = int(retries)
				if retries < 0 or retries > 10:
					self.errorLog(u"Retries value \"" + str(retries) + "\" is outside of the range of 0 to 10. Using default of 0 retry attempts.")
					retries = 0
			except ValueError:
				self.errorLog(u"Retries value \"" + str(retries) + "\" is not a valid number between 0 and 10. Using default of 0 retry attempts.")
				retries = 0
			self.debugLog("Retries: " + str(retries))
		
		# Convert the brightness to a 4-bit binary string.
		binaryBrightness = "0b0"	# Start out with 0 (off)
		# If the requested brightness is greater than 0 but less
		#   than the minimum brightness of 12%, set the requested
		#   brightness to 12%.
		if brightness > 0 and brightness < 12:
			brightness = 12
		# Use the brightness value closest to one of the 16 acceptable values.
		thisDiff = 0
		nextDiff = 0
		self.debugLog("Selecting closest brightness percentage.")
		for i, presetBrightness in enumerate(presetBrightnesses):
			thisDiff = presetBrightness - brightness
			try:
				nextDiff = brightness - presetBrightnesses[i + 1]
			except IndexError:
				nextDiff = brightness
			if nextDiff > thisDiff:
				brightness = presetBrightness
				binaryBrightness = bin(15 - i)
				# (Use 15 - i because the presetBrightnesses list order is backwards).
				break
		# If binaryBrightness is zero, use the turn off command instaed.
		if binaryBrightness == "0b0":
			command1 = "0x2f"
			# Correct the brightness value as well.
			brightness = 0
		
		# Convert the ramp rate time to a 4-bit binary string.
		binaryRate = ""
		thisDiff = 0
		nextDiff = 0
		# Use for logging.
		logRate = rate
		
		# If the specified rate is not one of the preset rates accepted
		#   by INSTEON dimmers, set the rate to the nearest preset time.
		self.debugLog("Selecting closest ramp rate time.")
		for i, presetRate in enumerate(presetRates):
			thisDiff = presetRate - rate
			try:
				nextDiff = rate - presetRates[i + 1]
			except IndexError:
				nextDiff = rate
			if nextDiff > thisDiff:
				logRate = presetRate
				binaryRate = bin(i)
				break
		# Catch if rate was zero and set it to 0.1.
		if rate == 0:
			rate = 0.1
			logRate = 0.1
			binaryRate = bin(15)	# (minimum ramp rate)
		
		# The binaryRate binary value must be exactly 4 digits long, so add
		#   leading zeros if needed.
		self.debugLog("Ramp rate to display (logRate): " + str(logRate) + ", ramp rate in binary (binaryRate): " + str(binaryRate))
		binaryRate = binaryRate[2:]		# Removes leading "0b" string.
		if len(binaryRate) < 4:
			for digit in range(1, 4 - len(binaryRate) + 1):
				binaryRate = "0" + binaryRate
		
		# Now combine the binaryBrightness and binaryRate into a single byte,
		#   convert that to an integer.
		command2 = int(binaryBrightness[2:] + binaryRate, 2)
		
		self.debugLog("Combining binary brightness (binaryBrightness " + str(binaryBrightness) + ") and binary rate (binaryRate " + str(binaryRate) + "): " + binaryBrightness[2:] + binaryRate + ". Resulting hex commands: (1) " + str(command1) + ", (2) " + str(hex(command2)))
		
		# Send the commands.
		# Convert the commands to integers.
		command1 = int(command1, 16)
		# command2 is already an integer.
		self.debugLog("Adding standard INSTEON command " + str(command1) + " " + str(command2) + " (integer format) to the command queue for device \"" + device.name + "\".")
		self.commandQueue.append([device.id, command1, command2, brightness, logRate, delay, retries])
		
		self.debugLog("setBrightness call complete.")
		
		return None
		
		# -- ALTERNATE CODE THAT DOESN'T USE THREADS BELOW --
		# If this is an i2CS device, use extended commands.
		# if isDeviceI2CS:
		#	message = [command1, command2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
		#	reply = indigo.insteon.sendRawExtended(device.address, message, waitUntilAck=True, suppressLogging=True)
		# else:
		#	message = [command1, command2]
		#	reply = indigo.insteon.sendRaw(device.address, message, waitUntilAck=True, suppressLogging=True)

	# Brighten by 1 Step
	########################################
	def brightenOneStep(self, action):
		self.debugLog("brightenOneStep called.")
		
		command1 = "0x15"	# Brighten by 1 step (there 32 steps in all)
		command2 = "0x00"	# (value ignored, but required)
		# Convert the commands to integers.
		command1 = int(command1, 16)
		command2 = int(command2, 16)
		
		device = indigo.devices[action.deviceId]
		self.debugLog(u"Device: " + device.name + u" (ID " + str(device.id) + u"). Device info:\n" + str(device))
		
		# Send the command right away (don't queue it).
		message = [command1, command2]
		try:
			reply = indigo.insteon.sendRaw(device.address, message, waitUntilAck=True, suppressLogging=True)
			self.debugLog(u"    reply: " + str(reply))
			indigo.server.log("\"" + device.name + u"\" brighten 1 step", u"Sent INSTEON")
		except Exception, e:
			errorMessage = u"\"" + device.name + u"\" brighten 1 step; send failed (no acknowledgment)."
			indigo.server.log(errorMessage, u"INSTEON Error", isError=True)
		try:
			reply = indigo.insteon.sendStatusRequest(device.address, waitUntilAck=True)
			self.debugLog(u"    status request reply: " + str(reply))
		except Exception, e:
			# No need to take special logging action or anything.
			return
			
	# Dim by 1 Step
	########################################
	def dimOneStep(self, action):
		self.debugLog("dimOneStep called.")
		
		command1 = "0x16"	# Dim by 1 step (there 32 steps in all)
		command2 = "0x00"	# (value ignored, but required)
		# Convert the commands to integers.
		command1 = int(command1, 16)
		command2 = int(command2, 16)
		
		device = indigo.devices[action.deviceId]
		self.debugLog(u"Device: " + device.name + u" (ID " + str(device.id) + u"). Device info:\n" + str(device))
		
		# Send the command right away (don't queue it).
		message = [command1, command2]
		try:
			reply = indigo.insteon.sendRaw(device.address, message, waitUntilAck=True, suppressLogging=True)
			self.debugLog(u"    reply: " + str(reply))
			indigo.server.log("\"" + device.name + u"\" dim 1 step", u"Sent INSTEON")
		except Exception, e:
			errorMessage = u"\"" + device.name + u"\" dim 1 step; send failed (no acknowledgment)."
			indigo.server.log(errorMessage, u"INSTEON Error", isError=True)
		try:
			reply = indigo.insteon.sendStatusRequest(device.address, waitUntilAck=True)
			self.debugLog(u"    status request reply: " + str(reply))
		except Exception, e:
			# No need to take special logging action or anything.
			return
			
	# Start Brightening
	########################################
	def startBrightening(self, action):
		self.debugLog("startBrightening called.")
		
		command1 = "0x17"	# Start manual light change command.
		command2 = "0x01"	# Up (brightening) direction.
		# Convert the commands to integers.
		command1 = int(command1, 16)
		command2 = int(command2, 16)
		
		device = indigo.devices[action.deviceId]
		self.debugLog(u"Device: " + device.name + u" (ID " + str(device.id) + u"). Device info:\n" + str(device))

		# Send the command right away (don't queue it).
		message = [command1, command2]
		try:
			reply = indigo.insteon.sendRaw(device.address, message, waitUntilAck=True, suppressLogging=True)
			self.debugLog(u"    reply: " + str(reply))
			indigo.server.log(u"\"" + device.name + u"\" start brighten", u"Sent INSTEON")
		except Exception, e:
			errorMessage = u"\"" + device.name + u"\" start brighten; send failed (no acknowledgment)."
			indigo.server.log(errorMessage, u"INSTEON Error", isError=True)
		
	# Start Dimming
	########################################
	def startDimming(self, action):
		self.debugLog("startDimming called.")
		
		command1 = "0x17"	# Start manual light change command.
		command2 = "0x00"	# Down (dimming) direction.
		# Convert the commands to integers.
		command1 = int(command1, 16)
		command2 = int(command2, 16)
		
		device = indigo.devices[action.deviceId]
		self.debugLog(u"Device: " + device.name + u" (ID " + str(device.id) + u"). Device info:\n" + str(device))

		# Send the command right away (don't queue it).
		message = [command1, command2]
		try:
			reply = indigo.insteon.sendRaw(device.address, message, waitUntilAck=True, suppressLogging=True)
			self.debugLog(u"    reply: " + str(reply))
			indigo.server.log(u"\"" + device.name + u"\" start dim", u"Sent INSTEON")
		except Exception, e:
			errorMessage = u"\"" + device.name + u"\" start dim; send failed (no acknowledgment)."
			indigo.server.log(errorMessage, u"INSTEON Error", isError=True)
		
	# Stop Brightening/Dimming
	########################################
	def stopBrighteningDimming(self, action):
		self.debugLog("stopBrighteningDimming called.")
		
		command1 = "0x18"	# Stop manual light change command.
		command2 = "0x00"	# (value ignored, but required)
		# Convert the commands to integers.
		command1 = int(command1, 16)
		command2 = int(command2, 16)
		
		device = indigo.devices[action.deviceId]
		self.debugLog(u"Device: " + device.name + u" (ID " + str(device.id) + u"). Device info:\n" + str(device))
		
		# Send the command right away (don't queue it).
		message = [command1, command2]
		try:
			reply = indigo.insteon.sendRaw(device.address, message, waitUntilAck=True, suppressLogging=True)
			self.debugLog(u"    reply: " + str(reply))
			indigo.server.log("\"" + device.name + u"\" stop brighten/dim", u"Sent INSTEON")
		except Exception, e:
			errorMessage = u"\"" + device.name + u"\" stop brighten/dim; send failed (no acknowledgment)."
			indigo.server.log(errorMessage, u"INSTEON Error", isError=True)
		try:
			reply = indigo.insteon.sendStatusRequest(device.address, waitUntilAck=True)
			self.debugLog(u"    status request reply: " + str(reply))
		except Exception, e:
			# No need to take special logging action or anything.
			return

